# Copyright 2014 MINES ParisTech
#
# This file is part of LinPy.
#
# LinPy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# LinPy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LinPy.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from ..domains import *
from ..linexprs import Symbol, symbols
from ..polyhedra import *


class TestDomain(unittest.TestCase):

    def setUp(self):
        x, y = symbols('x y')
        self.square1 = Polyhedron(inequalities=[x, 2 - x, y, 2 - y])
        self.square2 = Polyhedron(inequalities=[x - 1, 3 - x , y - 1, 3 - y]) #correct representation
        self.square3 = Polyhedron(inequalities=[x, 3 - x, y, 3 - y])
        self.square4 = Polyhedron(inequalities=[x - 1, 2 - x, y - 1, 2 - y])
        self.square5 = Polyhedron(inequalities=[x - 3, 6 - x, y - 3, 6 -y])
        self.square6 = Polyhedron(equalities=[3 - y], inequalities=[x - 1, 3 - x, y - 1])
        self.unbound_poly = Polyhedron(inequalities=[x, 3 - x, y])
        self.universe = Polyhedron([])
        self.empty = Empty
        self.disjoint = And(Ge(x, 0), Ge(-x + 2, 0), Ge(y, 0), Ge(-y + 2, 0))
        self.complement = Or(Ge(-x - 1, 0), Ge(x - 3, 0), And(Ge(x, 0), Ge(-x + 2, 0), Ge(-y - 1, 0)), And(Ge(x, 0), Ge(-x + 2, 0), Ge(y - 3, 0)))
        self.hull = And(Ge(x, 0), Ge(-x + 2, 0), Ge(y, 0), Ge(-y + 2, 0))
        self.dropped = And(Ge(y, 0), Ge(-y + 2, 0))
        self.intersection = And(Ge(x - 1, 0), Ge(-x + 2, 0), Ge(y - 1, 0), Ge(-y + 2, 0))
        self.union = Or(And(Ge(x, 0), Ge(-x + 2, 0), Ge(y, 0), Ge(-y + 2, 0)), And(Ge(x - 1, 0), Ge(-x + 3, 0), Ge(y - 1, 0), Ge(-y + 3, 0)))
        self.sum1 = Or(And(Ge(x, 0), Ge(-x + 2, 0), Ge(y, 0), Ge(-y + 2, 0)), And(Ge(x - 1, 0), Ge(-x + 3, 0), Ge(y - 1, 0), Ge(-y + 3, 0)))
        self.sum2 =And(Ge(x, 0), Ge(y, 0), Ge(-y + 3, 0), Ge(-x + 3, 0), Ge(x - y + 2, 0), Ge(-x + y + 2, 0))
        self.difference1 = Or(And(Eq(x - 3, 0), Ge(y - 1, 0), Ge(-y + 3, 0)), And(Eq(y - 3, 0), Ge(x - 1, 0), Ge(-x + 2, 0)))
        self.difference2 = And(Ge(x + y - 4, 0), Ge(-x + 3, 0), Ge(-y + 3, 0))
        self.lexmin = And(Eq(y, 0), Eq(x, 0))
        self.lexmax = And(Eq(y - 2, 0), Eq(x - 2, 0))

    def test_new(self):
        with self.assertRaises(TypeError):
            Polyhedron(1)

    def test_disjoint(self):
        self.assertEqual(self.square1.disjoint(), self.disjoint)
        self.assertEqual(self.empty.disjoint(), Empty)
        self.assertEqual(self.universe.disjoint(), self.universe)

    def test_isempty(self):
        self.assertFalse(self.square1.isempty())
        self.assertTrue(self.empty.isempty())
        self.assertFalse(self.universe.isempty())

    def test_isuniverse(self):
        self.assertFalse(self.square1.isuniverse())
        self.assertTrue(self.universe.isuniverse())

    def test_isbounded(self):
        self.assertTrue(self.square1.isbounded())
        self.assertFalse(self.unbound_poly.isbounded())

    def test_eq(self):
        self.assertTrue(self.square1 == self.square1)
        self.assertFalse(self.square1 == self.square2)
        self.assertFalse(self.empty == self.universe)

    def test_isdisjoint(self):
        self.assertFalse(self.square1.isdisjoint(self.square2))
        self.assertFalse(self.universe.isdisjoint(self.square1))
        self.assertTrue(self.square1.isdisjoint(self.square5))
        self.assertTrue(self.empty.isdisjoint(self.square1))

    def test_issubset(self):
        self.assertTrue(self.square4.issubset(self.unbound_poly))
        self.assertFalse(self.square1.issubset(self.square2))
        self.assertTrue(self.square1.issubset(self.universe))
        self.assertTrue(self.empty.issubset(self.square1))

    def test_le(self):
        self.assertTrue(self.square4 <= self.square3)
        self.assertFalse(self.square3 <= self.square4)
        self.assertTrue(self.empty <= self.square1)
        self.assertTrue(self.square1 <= self.universe)

    def test_lt(self):
        self.assertTrue(self.square4 < self.square3)
        self.assertFalse(self.square3 < self.square4)
        self.assertTrue(self.empty < self.square1)
        self.assertTrue(self.square1 < self.universe)

    def test_complement(self):
        self.assertEqual(~self.square1, self.complement)
        self.assertEqual(~self.universe, Empty)
        self.assertEqual(~self.empty, self.universe)

    def test_aspolyhedron(self):
        self.assertEqual(self.square1.aspolyhedron(), self.hull)
        self.assertEqual(self.universe.aspolyhedron(), self.universe)
        self.assertEqual(self.empty.aspolyhedron(), self.empty)

    def test_project(self):
        self.assertEqual(self.square1.project(symbols('x')), self.dropped)
        self.assertEqual(self.square1.project(symbols('x y')), self.universe)
        self.assertEqual(self.universe.project([]), self.universe)
        self.assertEqual(self.empty.project([]), Empty)

    def test_sample(self):
        self.assertEqual(self.square6.sample(), {Symbol('x'): 1, Symbol('y'): 3})
        with self.assertRaises(ValueError):
            self.empty.sample()
        self.assertEqual(self.universe.sample(), {})

    def test_intersection(self):
        self.assertEqual(self.square1.intersection(self.square2), self.intersection)

    def test_and(self):
        self.assertEqual(self.square2 & self.square1, self.intersection)
        self.assertEqual(self.square1 & self.universe, self.square1)
        self.assertEqual(self.empty & self.square1, Empty)
        self.assertEqual(self.universe & self.universe, self.universe)
        self.assertEqual(self.universe & self.empty, Empty)
        self.assertEqual(self.empty & self.empty, Empty)

    def test_union(self):
        self.assertEqual(self.square1.union(self.square2), self.union)
        self.assertEqual(self.square1.union(self.empty), self.square1)
        self.assertEqual(self.square1.union(self.universe), self.universe)
        self.assertEqual(self.universe.union(self.universe), self.universe)
        self.assertEqual(self.empty.union(self.empty), self.empty)

    def test_or(self):
        self.assertEqual(self.square1 | self.square2, self.union)

    def test_add(self):
        self.assertEqual(self.square2 + self.square1, self.sum1)
        self.assertEqual(Polyhedron(self.square1 + self.square2), self.sum2)
        self.assertEqual(self.universe + self.square1, self.universe)
        self.assertEqual(self.empty + self.square1, self.square1)
        self.assertEqual(self.universe + self.universe, self.universe)

    def test_difference(self):
        self.assertEqual(self.square2 - self.square1, self.difference1)
        self.assertEqual(Polyhedron(self.square2 - self.square1), self.difference2)
        self.assertEqual(self.square2 - self.square2, Empty)
        self.assertEqual(self.universe - self.universe, Empty)

    def test_lexmin(self):
        self.assertEqual(self.square1.lexmin(), self.lexmin)
        self.assertEqual(self.universe.lexmin(), self.universe)
        self.assertEqual(self.empty.lexmin(), Empty)

    def test_lexmax(self):
        self.assertEqual(self.square1.lexmax(), self.lexmax)
        self.assertEqual(self.universe.lexmax(), self.universe)
        self.assertEqual(self.empty.lexmax(), Empty)
